/*
 * Constable security daemon for Medusa DS9 security system
 */


#define __KERNEL__
#include <asm/types.h>
#undef __KERNEL__

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>
#include <string.h>
#include <sched.h>

#include "constable.h"

#define	BUFLEN	10240

char *me=NULL;
int medfd;

static int null_mod(	struct medusa_packet *msg, char *data,
			struct medusa_packet *ans, char **outdata )
{	return(0);	/* not for me */
}

struct c_module modules={NULL,null_mod};

int add_module( int(*func)(struct medusa_packet*,char *,
		struct medusa_packet *,char **) )
{ struct c_module *new;
	if( (new=malloc(sizeof(struct c_module)))==NULL )
		fatal(OUTMEM);
	new->func=func;
	new->next=modules.next;
	modules.next=new;
	return(0);
}

char *med_cmd( struct medusa_packet *msg, char *data )
{ static char buf[BUFLEN];
	msg->id=0;		
	msg->answer=MED_OK;
	if( data!=NULL && msg->data_len<=0 )
		msg->data_len=strlen(data)+1;
	if( write(medfd,msg,sizeof(struct medusa_packet))
			!=sizeof(struct medusa_packet) )
	{	fprintf(stderr, "%s: CMD write error\n", me);
		msg->answer=MED_ERR;
		return(NULL);
	}
	if( data!=NULL )
	{	if( write(medfd,data,msg->data_len)!=msg->data_len )
		{	fprintf(stderr, "%s: CMD write data error\n", me);
			msg->answer=MED_ERR;
			return(NULL);
		}
	}
	if( read(medfd,msg,sizeof(struct medusa_packet))
			!=sizeof(struct medusa_packet) )
	{	fprintf(stderr, "%s: CMD read error\n", me);
		msg->answer=MED_ERR;
		return(NULL);
	}
	if( msg->id!=0 )
	{	fprintf(stderr, "%s: Comunication error!\n", me);
		close(medfd);
		exit(-1);
	}
	if( msg->data_len>0 )
	{	if( msg->data_len >= BUFLEN )
		{	fprintf(stderr, "%s: Out of memory!\n", me);
			/* FIXME! tell kernel to let it be */
			close(medfd);
			exit(-1);
		}
		if( read(medfd,buf,msg->data_len)!=msg->data_len )
		{	fprintf(stderr, "%s: CMD read data error\n", me);
			msg->answer=MED_ERR;
			return(NULL);
		}
		buf[msg->data_len]=0;
		return(buf);
	}
	return(NULL);
}

int main( int argc, char *argv[] )
{ int r;
  struct c_module *mod;
  struct medusa_packet msg,ans;
  char *data_from_kernel, *data_to_kernel;
  struct sched_param schedpar;
  struct stat statbuf;
  char *devfile = "/dev/medusa", *conffile = "/usr/local/etc/constable.conf";

  char bufer[BUFLEN];
	me=argv && argv[0] ? argv[0] : "constable";
	if(argc > 1) {
		if(!strncmp(argv[1], "--help", 6)) {
			printf("Usage: %s [medusa-character device] [configuration file]\n",me);
			return(0);
		}
		devfile = argv[1];
	}
	if(argc > 2) {
		conffile = argv[2];
	}
	if( modules_init(conffile) < 0 )	return(-1);

#ifndef DEBUG
	switch (fork()) {
		case 0:
			setsid();
			schedpar.sched_priority = sched_get_priority_max(SCHED_FIFO);
			sched_setscheduler(
				0,
				SCHED_FIFO,
				&schedpar
			);
			break;
		case -1:
			return -1;
		default:
			return 0;
	}
#endif
	if(stat(devfile, &statbuf) != 0) {
		fprintf(stderr, "%s: Can't access %s\n",me, devfile);
		return(-1);
	}
	if(!S_ISCHR(statbuf.st_mode)) {
		fprintf(stderr, "%s: %s isn't a character device\n",me, devfile);
		return(-1);
	}
	if( (medfd=open(devfile,O_RDWR))<0 )
	{	fprintf(stderr, "%s: Can't open %s\n",me, devfile);
		return(-1);
	}
	msg.id = 0;		/* send supported communcation protocol */
	msg.cmd = MED_VERSION;	/* version number */
	msg.u.r_cmd.arg[0] = 1;	/* only one version is now supported :-) */
	msg.data_len = 0;
	med_cmd( &msg, NULL );
	if (msg.answer == MED_ERR) {
		fprintf(stderr, "%s: Usupported version of communication protocol\n",me);
		close(medfd);
		return(-1);
	}
	while(1)
	{
		fd_set in;
		FD_ZERO(&in);
		FD_SET(medfd,&in);
		select(medfd+1,&in,NULL,NULL,NULL);
		if( !FD_ISSET(medfd,&in) ) fprintf(stderr, "%s: Read error!\n", me);
		if( read(medfd,&msg,sizeof(msg))!=sizeof(msg) )
		{	fprintf(stderr, "%s: Read error!\n", me);
			close(medfd);
			return(-1);
		}
		ans.id=msg.id;
		ans.cmd=msg.cmd;
		ans.answer=MED_ERR;
		if( msg.cmd== 1000 )	break;
		/* get the data if available */
		if( msg.data_len>0 )
		{	if( msg.data_len >= BUFLEN )
			{	fprintf(stderr, "%s: Out of memory!\n", me);
				/* FIXME! tell kernel to let it be */ 
				close(medfd);
				return(-1);
			}
			data_from_kernel=bufer;
			if( read(medfd,data_from_kernel,msg.data_len)!=msg.data_len )
			{	fprintf(stderr, "%s: Read error!\n", me);
				close(medfd);
				return(-1);
			}
			data_from_kernel[msg.data_len]=0;
		}
		else	data_from_kernel=NULL;
		data_to_kernel=NULL;
//		if( msg.cmd==MED_IACT )		/* to make medinf valid */
//		{   ans.u.a_iact.proc.pid=0;
//		    if( msg.u.r_iact.inode.ino!=0 && msg.u.r_iact.inode.dev!=0
//			    && !med_c_valid( &(msg.u.r_iact.inode.mi) ) )
//		    { files_t *f;
//printf("BUUU 1\n");
//			f=files_find_ino( root, msg.u.r_iact.inode.dev,
//					  msg.u.r_iact.inode.ino );
//			msg.u.r_iact.inode.mi.cinfo=(u_long)f;
//		    }
//		}
		mod=&modules;
		while( mod!=NULL )
		{	if( (r=mod->func(&msg,data_from_kernel,
					 &ans,&data_to_kernel))!=0 )
			{	if( r<0 )
					ans.answer=MED_ERR;
				break;
			}
			mod=mod->next;
		}
		if( mod==NULL )
			ans.answer=MED_ERR;
		if( ans.answer!=MED_ERR && data_to_kernel!=NULL )
			ans.data_len=strlen(data_to_kernel)+1;
		else	ans.data_len=0;
		if( write(medfd,&ans,sizeof(ans))!=sizeof(ans) )
		{	fprintf(stderr, "%s: Write error!\n", me);
			close(medfd);
			return(-1);
		}
		if( ans.answer!=MED_ERR && data_to_kernel!=NULL )
		{	if( write(medfd,data_to_kernel,ans.data_len)!=ans.data_len )
			{	fprintf(stderr, "%s: Data Write error!\n", me);
				close(medfd);
				return(-1);
			}
		}
	}
	close(medfd);
	return(0);
}

